"""CycloMonitor locales (IBTrACS module)"""

import locale
import json
import pathlib
import os
import sys as _sys
import logging as _logging

IBTRACS_DB_NOT_FOUND = "IBTRACS_DB_NOT_FOUND"
CLASS_TD = "CLASS_TD"
CLASS_SD = "CLASS_SD"
CLASS_TS = "CLASS_TS"
CLASS_SS = "CLASS_SS"
CLASS_HU = "CLASS_HU"
CLASS_TY = "CLASS_TY"
CLASS_STY = "CLASS_STY"
CLASS_TC = "CLASS_TC"
CLASS_STC = "CLASS_STC"
QUERY_REPR = "QUERY_REPR"
ERROR_ILLEGAL_UPDATE_MODE = "ERROR_ILLEGAL_UPDATE_MODE"
IBTRACS_UPDATE_LAST3 = "IBTRACS_UPDATE_LAST3"
IBTRACS_UPDATE_ALL = "IBTRACS_UPDATE_ALL"
IBTRACS_UPDATE_FULL = "IBTRACS_UPDATE_FULL"
IBTRACS_GETTING_DATA = "IBTRACS_GETTING_DATA"
ERROR_IBTRACS_UPDATE_FAILURE = "ERROR_IBTRACS_UPDATE_FAILURE"
ERROR_MISSING_IBTRACS_DB = "ERROR_MISSING_IBTRACS_DB"
ERROR_INVALID_TABLE = "ERROR_INVALID_TABLE"
ERROR_INVALID_BASIN = "ERROR_INVALID_BASIN"
ERROR_INVALID_SEASON = "ERROR_INVALID_SEASON"
ERROR_NO_PARAMS = "ERROR_NO_PARAMS"
IBTRACS_CONDS = "IBTRACS_CONDS"
IBTRACS_UPDATE_SUCCESS = "IBTRACS_UPDATE_SUCCESS"

_log = _logging.getLogger(__name__)
locale.setlocale(locale.LC_ALL, "")
_PATH = os.path.dirname(os.path.realpath(__file__))


def set_locale(lang="C"):
    """Set the locale to `lang`. Returns the language that was set."""
    json_file = pathlib.Path(f"{_PATH}/{lang}.json")
    try:
        with json_file.open() as f:
            lc = json.load(f)
    except OSError:
        _log.exception(f"Locale {lang} not found. Falling back to default locale (C).")
        lang = "C"
        with open(f"{_PATH}/C.json") as f:
            lc = json.load(f)

    globals().update({k: v for k, v in lc.items()})
    return lang


def locale_init():
    global lang
    if _sys.platform.startswith("win32"):
        import ctypes

        windll = ctypes.windll.kernel32
        lang = locale.windows_locale[windll.GetUserDefaultUILanguage()]
    else:
        lang, _ = locale.getlocale()
    lang = set_locale(lang)


locale_init()



def get_dir(deg: float) -> str:
    if deg < 0:
        return ""
    deg %= 360
    if deg > 348.75 or deg <= 11.25:
        return "N"
    elif deg > 11.25 and deg <= 33.75:
        return "NNE"
    elif deg > 33.75 and deg <= 56.25:
        return "NE"
    elif deg > 56.25 and deg <= 78.75:
        return "ENE"
    elif deg > 78.75 and deg <= 101.25:
        return "E"
    elif deg > 101.25 and deg <= 123.75:
        return "ESE"
    elif deg > 123.75 and deg <= 146.25:
        return "SE"
    elif deg > 146.25 and deg <= 168.75:
        return "SSE"
    elif deg > 168.75 and deg <= 191.25:
        return "S"
    elif deg > 191.25 and deg <= 213.75:
        return "SSW"
    elif deg > 213.75 and deg <= 236.25:
        return "SW"
    elif deg > 236.25 and deg <= 258.75:
        return "WSW"
    elif deg > 258.75 and deg <= 281.25:
        return "W"
    elif deg > 281.25 and deg <= 303.75:
        return "WNW"
    elif deg > 303.75 and deg <= 326.25:
        return "NW"
    elif deg > 326.25 and deg <= 348.75:
        return "NNW"



"""CycloMonitor CLI"""

import asyncio
import datetime
import re
from sys import exit, version_info
from .atcf import *
from . import errors
from .ibtracs import *
from . import locales
from .dir_calc import get_dir
from .locales import *
from io import StringIO
from os import chdir  # for your convenience
from typing import Callable, Awaitable, Generator

COPYRIGHT_NOTICE = """\
CycloMonitor - ATCF and IBTrACS wrapper for Discord
Copyright (c) 2023 Virtual Nate

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
details.

For those hosting a copy of this bot, you should have received a copy of the
GNU Affero General Public License along with this program. If not, see
<https://www.gnu.org/licenses/>."""
KT_TO_MPH = 1.15077945
KT_TO_KMH = 1.852
# fmt: off
# Attributes that probably shouldn't be accessed from the CLI
PRIVATE_ATTRS = {
    "zip_longest", "ATCFError", "WrongData", "NoActiveStorms", "main",
    "dataclass", "Iterable", "Storm", "Query", "query_group", "varchar",
    "numeric", "bit", "StringIO", "Callable", "Awaitable", "Generator",
    "Internal", "PRIVATE_ATTRS", "log", "asyncio", "datetime", "logging",
    "aiohttp", "json", "sqlite3", "subprocess", "io", "re", "version_info",
    "Literal", "Tuple",
}
PRIVATE_ATTRS.update(
    attr for attr in dir() if not isinstance(globals()[attr], Callable)
)
PRIVATE_ATTRS.remove("KT_TO_MPH")
PRIVATE_ATTRS.remove("KT_TO_KMH")
CONSTANTS = {
    "CONSTANTS", "PRIVATE_ATTRS", "COPYRIGHT_NOTICE", "KT_TO_MPH", "KT_TO_KMH",
    "cyclones", "names", "timestamps", "lats", "longs", "basins", "winds",
    "pressures", "tc_classes", "lats_real", "longs_real", "movement_speeds",
    "movement_dirs", "None", "True", "False",
}
log = logging.getLogger(__name__)
cd = chdir
# fmt: on


class Internal:
    """
    Methods internal to the CLI.
    All methods in this class (except for __repr__) are static.
    """

    def __repr__(self):
        return CLI_PARSER_REPR

    @staticmethod
    def parse(line: str):
        # TODO: Support quoted arguments
        if not line:
            return ""
        args = line.split()
        kwargs = {
            arg.split("=")[0]: arg.split("=")[1] for arg in args[1:] if "=" in arg
        }
        # Separate args and kwargs
        to_delete = []
        for arg in args[1:]:
            if "=" in arg:
                to_delete.append(arg)
        for arg in to_delete:
            args.remove(arg)

        for index, arg in enumerate(args):
            if arg.startswith("$"):
                args[index] = globals().get(arg[1:], "")

            if index == 0:
                continue
            if "." in arg:
                try:
                    args[index] = float(arg)
                except ValueError:
                    pass
            else:
                try:
                    args[index] = int(arg)
                except ValueError:
                    pass
        for k, v in kwargs.items():
            if v.startswith("$"):
                kwargs[k] = globals().get(v[1:], "")

            if "." in v:
                try:
                    kwargs[k] = float(v)
                except ValueError:
                    pass
            else:
                try:
                    kwargs[k] = int(v)
                except ValueError:
                    pass

        if args[0] in globals() and args[0] not in PRIVATE_ATTRS:
            command = globals()[args[0]]
            if isinstance(command, Callable):
                try:
                    out = command(*args[1:], **kwargs)
                except Exception as e:
                    if e.args:
                        return f"{type(e).__name__}: {e}"
                    else:
                        return type(e).__name__

                if isinstance(out, Awaitable):
                    try:
                        out = asyncio.run(out)
                    except Exception as e:
                        if e.args:
                            return f"{type(e).__name__}: {e}"
                        else:
                            return type(e).__name__

                if out is not None and not isinstance(out, (str, Storm, Generator)):
                    return repr(out)
                elif out is None:
                    return ""
                # almost certainly the output of ibtracs.get_storm
                elif isinstance(out, Storm):
                    if out.name == "NOT_NAMED":
                        header = CM_UNNAMED_STORM.format(out.nature())
                    else:
                        header = f"{out.nature()} {out.name}"
                    return CM_PAST_STORM_INFO.format(
                        header,
                        out.season,
                        out.basin,
                        out.peak_winds,
                        out.peak_pres,
                        out.time_of_peak,
                        out.atcf_id,
                        out.best_track_id,
                    )
                elif isinstance(out, Generator):
                    out_temp = StringIO()
                    if args[0] == "get_storm":
                        out_temp.write(CM_MULTIPLE_STORMS)
                    out_temp.writelines([f"{s}\n" for s in out])
                    out = out_temp.getvalue()
                    out_temp.close()
                return out
            elif isinstance(command, str):
                return command
            else:
                return repr(command)
        else:
            return CLI_SYMBOL_NOT_FOUND.format(args[0])

    @staticmethod
    def get_nature_name(name, wind, tc_class):
        # see cyclomonitor.update_guild for more info
        if name == "INVEST" and (not (tc_class == "SD" or tc_class == "SS")):
            tc_class = CLASS_AOI
        if tc_class == "EX":
            if not name == "INVEST":
                tc_class = CLASS_PTC
        elif tc_class == "LO" or tc_class == "INVEST":
            if not name == "INVEST":
                tc_class = CLASS_PTC
        elif tc_class == "DB" or tc_class == "WV":
            if not name == "INVEST":
                tc_class = CLASS_RL
        elif wind < 35:
            if not (tc_class == "SD" or name == "INVEST"):
                tc_class = CLASS_TD
            elif name == "INVEST" and not tc_class == "SD":
                pass
            else:
                tc_class = CLASS_SD
        elif wind > 34 and wind < 65:
            if not (tc_class == "SS" or name == "INVEST"):
                tc_class = CLASS_TS
            elif name == "INVEST" and not tc_class == "SS":
                pass
            else:
                tc_class = CLASS_SS
        else:
            if basin == "ATL" or basin == "EPAC" or basin == "CPAC":
                if wind < 100:
                    tc_class = CLASS_HU
                else:
                    tc_class = CLASS_MH
            elif basin == "WPAC":
                if wind < 130:
                    tc_class = CLASS_TY
                else:
                    tc_class = CLASS_STY
            else:
                tc_class = CLASS_CY
        return tc_class


def echo(*args):
    """Print a string to the screen."""
    return " ".join(args)


def copyright(*args):
    """Get the copyright notice."""
    return COPYRIGHT_NOTICE


def commands(*args):
    """Get the list of commands usable in the CLI."""
    return (
        key
        for key, value in globals().items()
        if isinstance(value, Callable) and key not in PRIVATE_ATTRS
    )


def attrs(*args):
    """Get the list of variables. Includes constants."""
    return (
        key
        for key, value in globals().items()
        if not (isinstance(value, Callable) or key in PRIVATE_ATTRS)
    )


def help(command_name="", *args):
    """Stop it. Get some help."""
    if command_name:
        try:
            command = globals()[command_name]
            if isinstance(command, Callable) and command_name not in PRIVATE_ATTRS:
                if command.__doc__ is not None:
                    return command.__doc__
                else:
                    return CLI_MISSING_DOCSTRING
            elif not isinstance(command, Callable):
                return CLI_IS_A_VAR.format(command_name)
            else:
                return CLI_SYMBOL_NOT_FOUND.format(command_name)
        except (AttributeError, KeyError):
            return CLI_SYMBOL_NOT_FOUND.format(command_name)
    else:
        return CLI_HELP


def set_var(var_name: str, value, *args):
    """Create var_name if it doesn't exist and set it to value.

    Paramaters:
    var_name - The name of the variable
    value - The value to set this variable to

    Precondition: var_name does not refer to a constant, command, or private
    attribute.
    """
    if var_name in CONSTANTS:
        return CLI_IS_A_CONSTANT
    elif var_name in PRIVATE_ATTRS or isinstance(globals().get(var_name), Callable):
        return CLI_ILLEGAL_NAME

    if args:
        value += f" {' '.join(args)}"

    globals()[var_name] = value
    return value


def active_storms(*args):
    """Get the list of active storms.
    Format is ID NAME.
    """
    return (f"{id} {name}" for id, name in zip(cyclones, names))


def present(name_or_id: str):
    """Present a TC in the ATCF data.

    Parameters:
    name_or_id - The TC name or identifier.
    """
    if not (cyclones and names):
        raise NoActiveStorms
    name_or_id = name_or_id.upper()
    if name_or_id in names:
        name = name_or_id
        index = names.index(name_or_id)
        id = cyclones[index]
    elif name_or_id in cyclones:
        id = name_or_id
        index = cyclones.index(name_or_id)
        name = names[index]
    else:
        raise ValueError(CLI_STORM_NOT_FOUND.format(name_or_id))

    tc_class = tc_classes[index]
    wind = winds[index]
    tc_class = Internal.get_nature_name(name, wind, tc_class)
    timestamp = (
        datetime.datetime.fromtimestamp(timestamps[index])
        .replace(tzinfo=datetime.timezone.utc)
        .isoformat()
    )
    if name == "INVEST":
        name = display_name = id
    else:
        display_name = f"{id} ({name})"
    lat = lats[index]
    long = longs[index]
    mph = round(wind * KT_TO_MPH / 5) * 5
    kmh = round(wind * KT_TO_KMH / 5) * 5
    pres = pressures[index]
    movement_speed = movement_speeds[index]
    movement_dir = get_dir(movement_dirs[index])
    if (not movement_dir) or (movement_speed) < 0:
        movement_str = NOT_AVAILABLE
    else:
        movement_mph = movement_speed * KT_TO_MPH
        movement_kmh = movement_speed * KT_TO_KMH
        movement_str = STORM_MOVEMENT.format(
            movement_dir, movement_speed, movement_mph, movement_kmh
        )
    return CM_STORM_INFO.format(
        "",
        tc_class,
        display_name,
        timestamp,
        name,
        lat,
        long,
        wind,
        mph,
        kmh,
        pres,
        movement_str,
    )


def cli():
    """The main function of the CLI."""
    print(CLI_STARTUP)
    if version_info.major == 3 and version_info.minor < 11:
        log.warning(
            CLI_OLD_PY_VERSION.format(f"{version_info.major}.{version_info.minor}")
        )
    while True:
        try:
            request = input("cyclomonitor> ")
        except EOFError:
            print("\n", end="")
            break
        except KeyboardInterrupt:
            print("\n", end="")
            continue
        except Exception:
            logging.exception(CLI_INTERNAL_ERROR)
            continue
        try:
            out = Internal.parse(request)
        except Exception:
            logging.exception(CLI_INTERNAL_ERROR)
            continue
        if out:
            if out.endswith("\n"):
                print(out, end="")
            else:
                print(out)


if __name__ == "__main__":
    logging.basicConfig(
        format="%(asctime)s.%(msecs)d %(name)s %(levelname)s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        level=logging.WARNING,
    )
    locale_init()
    cli()



# CycloMonitor Exceptions
# CycloMonitor Copyright (C) 2023 Nathaniel Greenwell
# This program comes with ABSOLUTELY NO WARRANTY; for details see main.py
class AutoUpdateError(Exception):
    pass


class LogRequested(Exception):
    pass



# uptime module
# CycloMonitor Copyright (C) 2023 Nathaniel Greenwell
# This program comes with ABSOLUTELY NO WARRANTY; for details see main.py
import time
import math

start_time = time.time()


def process_uptime():
    return time.time() - start_time


def process_uptime_human_readable():
    t_uptime = math.floor(process_uptime())
    days = math.floor(t_uptime / 86400)
    hours = math.floor((t_uptime % 86400) / 3600)
    minutes = math.floor((t_uptime % 3600) / 60)
    seconds = math.floor(t_uptime % 60)
    if days > 0:
        if days == 1:
            p_days = f"{days} day, "
        else:
            p_days = f"{days} days, "
    else:
        p_days = ""
    if hours > 0:
        if hours == 1:
            p_hours = f"{hours} hour, "
        else:
            p_hours = f"{hours} hours, "
    else:
        p_hours = ""
    if minutes > 0:
        if minutes == 1:
            p_minutes = f"{minutes} minute, "
        else:
            p_minutes = f"{minutes} minutes, "
    else:
        p_minutes = ""
    if seconds == 1:
        p_seconds = f"{seconds} second"
    else:
        p_seconds = f"{seconds} seconds"
    return f"{p_days}{p_hours}{p_minutes}{p_seconds}"



"""CycloMonitor locales"""

import locale
import json
import pathlib
import os
import sys as _sys
import logging as _logging

LOG_TENDO_NOT_FOUND = "LOG_TENDO_NOT_FOUND"
ERROR_ALREADY_RUNNING = "ERROR_ALREADY_RUNNING"
LOG_MONITOR_STOP = "LOG_MONITOR_STOP"
LOG_TIMESTAMP_COMPARISON = "LOG_TIMESTAMP_COMPARISON"
LOG_AUTO_UPDATE_BEGIN = "LOG_AUTO_UPDATE_BEGIN"
ERROR_AUTO_UPDATE_FAILED = "ERROR_AUTO_UPDATE_FAILED"
LOG_SUPPRESSED = "LOG_SUPPRESSED"
LOG_SUPPRESSED_TRY_2 = "LOG_SUPPRESSED_TRY_2"
CM_SUPPRESSED_MESSAGE = "CM_SUPPRESSED_MESSAGE"
NEXT_AUTO_UPDATE = "NEXT_AUTO_UPDATE"
CM_ERROR_WHILE_UPDATING = "CM_ERROR_WHILE_UPDATING"
CM_ATTACH_LOG = "CM_ATTACH_LOG"
ERROR_LOG_SEND_FAIL = "ERROR_LOG_SEND_FAIL"
LOG_NO_OWNER = "LOG_NO_OWNER"
CM_AUTO_UPDATE_FAILED_MESSAGE = "CM_AUTO_UPDATE_FAILED_MESSAGE"
LOG_IBTRACS_UPDATE_BEGIN = "LOG_IBTRACS_UPDATE_BEGIN"
LOG_IBTRACS_UPDATE_FAILED_ATTEMPT_1 = "LOG_IBTRACS_UPDATE_FAILED_ATTEMPT_1"
LOG_NEXT_ATTEMPT = "LOG_NEXT_ATTEMPT"
LOG_ATTEMPT_FAILED = "LOG_ATTEMPT_FAILED"
ERROR_IBTRACS_UPDATE_FAILED = "ERROR_IBTRACS_UPDATE_FAILED"
LOG_TRY_AGAIN = "LOG_TRY_AGAIN"
LOG_UPDATE_GUILD = "LOG_UPDATE_GUILD"
NOT_AVAILABLE = "NOT_AVAILABLE"
STORM_MOVEMENT = "STORM_MOVEMENT"
LOG_BASIN = "LOG_BASIN"
CLASS_AOI = "CLASS_AOI"
CLASS_PTC = "CLASS_PTC"
CLASS_RL = "CLASS_RL"
CLASS_TD = "CLASS_TD"
CLASS_SD = "CLASS_SD"
CLASS_TS = "CLASS_TS"
CLASS_SS = "CLASS_SS"
CLASS_HU = "CLASS_HU"
CLASS_MH = "CLASS_MH"
CLASS_TY = "CLASS_TY"
CLASS_STY = "CLASS_STY"
CLASS_CY = "CLASS_CY"
LOG_NEW_RECORD = "LOG_NEW_RECORD"
LOG_NO_RECORD = "LOG_NO_RECORD"
CM_STORM_INFO = "CM_STORM_INFO"
LOG_GUILD_UNAVAILABLE = "LOG_GUILD_UNAVAILABLE"
CM_NO_STORMS = "CM_NO_STORMS"
NO_AUTO_UPDATE = "NO_AUTO_UPDATE"
CM_NEXT_AUTO_UPDATE = "CM_NEXT_AUTO_UPDATE"
CM_MORE_INFO = "CM_MORE_INFO"
CM_WATCHING = "CM_WATCHING"
LOG_READY = "LOG_READY"
LOG_NEW_GUILD = "LOG_NEW_GUILD"
CM_GUILD_ADDED = "CM_GUILD_ADDED"
LOG_GUILD_REMOVED = "LOG_GUILD_REMOVED"
CM_NO_PERMISSION = "CM_NO_PERMISSION"
LOG_NO_PERMISSION = "LOG_NO_PERMISSION"
CM_NO_DM = "CM_NO_DM"
LOG_COMMAND_ERROR = "LOG_COMMAND_ERROR"
CM_COMMAND_ERROR = "CM_COMMAND_ERROR"
ERROR_CANNOT_RESPOND = "ERROR_CANNOT_RESPOND"
CM_PING = "CM_PING"
CM_PONG = "CM_PONG"
CM_SET_TRACKING_CHANNEL = "CM_SET_TRACKING_CHANNEL"
CM_CHANNEL_TO_USE = "CM_CHANNEL_TO_USE"
ERROR_NOT_A_TEXT_CHANNEL = "ERROR_NOT_A_TEXT_CHANNEL"
CM_SET_CHANNEL_SUCCESS = "CM_SET_CHANNEL_SUCCESS"
CM_CANNOT_SEND_MESSAGE = "CM_CANNOT_SEND_MESSAGE"
CM_UPDATE = "CM_UPDATE"
CM_UPDATE_SUCCESS = "CM_UPDATE_SUCCESS"
ERROR_NO_TRACKING_CHANNEL = "ERROR_NO_TRACKING_CHANNEL"
CM_UPDATE_ALT = "CM_UPDATE_ALT"
CM_SET_BASINS = "CM_SET_BASINS"
CM_NATL = "CM_NATL"
CM_EPAC = "CM_EPAC"
CM_CPAC = "CM_CPAC"
CM_WPAC = "CM_WPAC"
CM_NIO = "CM_NIO"
CM_SHEM = "CM_SHEM"
CM_BASINS_SAVED = "CM_BASINS_SAVED"
CM_UPDATE_ALL = "CM_UPDATE_ALL"
CM_UPDATE_ALL_ALT = "CM_UPDATE_ALL_ALT"
CM_ANNOUNCE_ALL = "CM_ANNOUNCE_ALL"
CM_TO_ANNOUNCE = "CM_TO_ANNOUNCE"
CM_ANNOUNCE_ALL_SUCCESS = "CM_ANNOUNCE_ALL_SUCCESS"
CM_ANNOUNCE_BASIN = "CM_ANNOUNCE_BASIN"
CM_BASIN_TO_ANNOUNCE = "CM_BASIN_TO_ANNOUNCE"
CM_BASIN_ANNOUNCEMENT = "CM_BASIN_ANNOUNCEMENT"
CM_ANNOUNCE_BASIN_SUCCESS = "CM_ANNOUNCE_BASIN_SUCCESS"
CM_ANNOUNCE_FILE = "CM_ANNOUNCE_FILE"
CM_TXT_FILE = "CM_TXT_FILE"
ERROR_NOT_A_TXT_FILE = "ERROR_NOT_A_TXT_FILE"
CM_INVITE = "CM_INVITE"
CM_INVITE_MESSAGE = "CM_INVITE_MESSAGE"
CM_STATISTICS_DESC = "CM_STATISTICS_DESC"
ERROR_NO_GLOBAL_VARS = "ERROR_NO_GLOBAL_VARS"
CM_STATISTICS = "CM_STATISTICS"
CM_YIKES = "CM_YIKES"
CM_INC_YIKES_COUNT = "CM_INC_YIKES_COUNT"
CM_YIKES_RESPONSE = "CM_YIKES_RESPONSE"
CM_GET_DATA = "CM_GET_DATA"
CM_GET_DATA_SUCCESS = "CM_GET_DATA_SUCCESS"
CM_GET_DATA_FAILED = "CM_GET_DATA_FAILED"
CM_GET_DATA_ALT = "CM_GET_DATA_ALT"
CM_ATCF_RESET = "CM_ATCF_RESET"
CM_ATCF_RESET_SUCCESS = "CM_ATCF_RESET_SUCCESS"
CM_GITHUB = "CM_GITHUB"
CM_GITHUB_RESPONSE = "CM_GITHUB_RESPONSE"
CM_COPYRIGHT = "CM_COPYRIGHT"
CM_RSMC_LIST = "CM_RSMC_LIST"
CM_RSMC_LIST_RESPONSE = "CM_RSMC_LIST_RESPONSE"
CM_GET_LOG = "CM_GET_LOG"
LOG_REQUESTED = "LOG_REQUESTED"
CM_LOG_SENT = "CM_LOG_SENT"
CM_SUSPEND_UPDATES = "CM_SUSPEND_UPDATES"
CM_SUSPEND_UPDATES_SUCCESS = "CM_SUSPEND_UPDATES_SUCCESS"
CM_UPDATES_ALREADY_SUSPENDED = "CM_UPDATES_ALREADY_SUSPENDED"
CM_RESUME_UPDATES = "CM_RESUME_UPDATES"
CM_RESUME_UPDATES_SUCCESS = "CM_RESUME_UPDATES_SUCCESS"
CM_UPDATES_ALREADY_RUNNING = "CM_UPDATES_ALREADY_RUNNING"
CM_FEEDBACK = "CM_FEEDBACK"
CM_FEEDBACK_TO_SEND = "CM_FEEDBACK_TO_SEND"
CM_FEEDBACK_RECEIVED = "CM_FEEDBACK_RECEIVED"
CM_FEEDBACK_SENT = "CM_FEEDBACK_SENT"
CM_NO_OWNER = "CM_NO_OWNER"
CM_GET_PAST_STORM = "CM_GET_PAST_STORM"
CM_PAST_STORM_NAME = "CM_PAST_STORM_NAME"
CM_PAST_STORM_SEASON = "CM_PAST_STORM_SEASON"
CM_PAST_STORM_BASIN = "CM_PAST_STORM_BASIN"
CM_PAST_STORM_ATCF = "CM_PAST_STORM_ATCF"
CM_PAST_STORM_SID = "CM_PAST_STORM_SID"
CM_PAST_STORM_TABLE = "CM_PAST_STORM_TABLE"
CM_WAIT_FOR_IBTRACS_UPDATE = "CM_WAIT_FOR_IBTRACS_UPDATE"
CM_SEARCHING = "CM_SEARCHING"
CM_ERROR = "CM_ERROR"
CM_MULTIPLE_STORMS = "CM_MULTIPLE_STORMS"
CM_UNNAMED_STORM = "CM_UNNAMED_STORM"
CM_UNKNOWN = "CM_UNKNOWN"
CM_PAST_STORM_INFO = "CM_PAST_STORM_INFO"
CM_NO_RESULTS = "CM_NO_RESULTS"
ERROR_WTF = "ERROR_WTF"
CM_SET_LANGUAGE = "CM_SET_LANGUAGE"
CM_LANG_TO_USE = "CM_LANG_TO_USE"
CM_SET_LANGUAGE_SUCCESS = "CM_SET_LANGUAGE_SUCCESS"
ATCF_PARSE_STORM = "ATCF_PARSE_STORM"
ATCF_ERROR_COL = "ATCF_ERROR_COL"
ATCF_WRONG_DATA = "ATCF_WRONG_DATA"
ATCF_NO_DATA = "ATCF_NO_DATA"
ERROR_HDYGH = "ERROR_HDYGH"
ATCF_GET_INTERP_FAILED = "ATCF_GET_INTERP_FAILED"
ATCF_USING_MAIN = "ATCF_USING_MAIN"
ATCF_USING_MAIN_FAILED = "ATCF_USING_MAIN_FAILED"
ATCF_USING_ALT = "ATCF_USING_ALT"
ERROR_TIMED_OUT = "ERROR_TIMED_OUT"
ERROR_ATCF_GET_DATA_FAILED = "ERROR_ATCF_GET_DATA_FAILED"
ERROR_GET_FORECAST_NO_PARAMS = "ERROR_GET_FORECAST_NO_PARAMS"
CM_GET_FORECAST = "CM_GET_FORECAST"
CM_NO_ACTIVE_STORMS = "CM_NO_ACTIVE_STORMS"
CM_CANNOT_FIND_STORM = "CM_CANNOT_FIND_STORM"
CM_IS_AN_INVEST = "CM_IS_AN_INVEST"
DESCRIPTION = "DESCRIPTION"
HELP_EPILOG = "HELP_EPILOG"
HELP_BOT = "HELP_BOT"
HELP_TOKEN = "HELP_TOKEN"
HELP_INTERACTIVE = "HELP_INTERACTIVE"
HELP_LOG_FILE = "HELP_LOG_FILE"
HELP_VERBOSE = "HELP_VERBOSE"
HELP_CONFIG = "HELP_CONFIG"
ERROR_BOT_AND_INTERACTIVE = "ERROR_BOT_AND_INTERACTIVE"
ERROR_NO_TOKEN = "ERROR_NO_TOKEN"
CM_SERVER = "CM_SERVER"
ERROR_PYCORD_MISSING = "ERROR_PYCORD_MISSING"
ERROR_TENDO_MISSING = "ERROR_TENDO_MISSING"
CLI_STARTUP = "CLI_STARTUP"
ERROR_LOG_FILE_AND_INTERACTIVE = "ERROR_LOG_FILE_AND_INTERACTIVE"
LOG_BOT_CONFIG_AND_INTERACTIVE = "LOG_BOT_CONFIG_AND_INTERACTIVE"
CLI_SYMBOL_NOT_FOUND = "CLI_SYMBOL_NOT_FOUND"
CLI_INTERNAL_ERROR = "CLI_INTERNAL_ERROR"
CLI_PARSER_REPR = "CLI_PARSER_REPR"
CLI_HELP = "CLI_HELP"
CLI_MISSING_DOCSTRING = "CLI_MISSING_DOCSTRING"
CLI_IS_A_VAR = "CLI_IS_A_VAR"
CLI_STORM_NOT_FOUND = "CLI_STORM_NOT_FOUND"
CLI_IS_A_CONSTANT = "CLI_IS_A_CONSTANT"
CLI_ILLEGAL_NAME = "CLI_ILLEGAL_NAME"
CLI_OLD_PY_VERSION = "CLI_OLD_PY_VERSION"

_log = _logging.getLogger(__name__)
locale.setlocale(locale.LC_ALL, "")
_PATH = os.path.dirname(os.path.realpath(__file__))


def set_locale(lang="C"):
    """Set the locale to `lang`. Returns the language that was set."""
    json_file = pathlib.Path(f"{_PATH}/{lang}.json")
    try:
        with json_file.open() as f:
            lc = json.load(f)
    except OSError:
        _log.exception(f"Locale {lang} not found. Falling back to default locale (C).")
        lang = "C"
        with open(f"{_PATH}/C.json") as f:
            lc = json.load(f)

    globals().update({k: v for k, v in lc.items()})
    return lang


def locale_init():
    global lang
    if _sys.platform.startswith("win32"):
        import ctypes

        windll = ctypes.windll.kernel32
        lang = locale.windows_locale[windll.GetUserDefaultUILanguage()]
    else:
        lang, _ = locale.getlocale()
    lang = set_locale(lang)


locale_init()



# CycloMonitor Copyright (C) 2023 Nathaniel Greenwell
# This program comes with ABSOLUTELY NO WARRANTY; for details see main.py
"""
CycloMonitor ATCF module

Classes:
ATCFError -- base exception for ATCF errors
WrongData -- exception for invalid data
Functions:
reset -- reset ATCF data
parse_storm -- parse ATCF data
load -- load ATCF data
get_data -- get ATCF data
get_data_alt -- get ATCF data (alt source)
"""
import datetime
import aiohttp
import asyncio
import json
import logging
from itertools import zip_longest
from .locales import *

# initalize variables
# For whatever reason I have yet to discover, the domain `www.nrlmry.navy.mil`
# causes requests from aiohttp to take much longer than expected, so I have to
# directly connect to its IP address.
URL = "https://199.9.2.143/tcdat/sectors/atcf_sector_file"
URL_INTERP = "https://199.9.2.143/tcdat/sectors/interp_sector_file"
URL_ALT = "https://api.knackwx.com/atcf/v2"
BASE_URL_NHC = (
    "https://www.nhc.noaa.gov/storm_graphics/{0}/{1}_5day_cone_with_line_and_wind.png"
)
BASE_URL_JTWC = "https://www.metoc.navy.mil/jtwc/products/{0}.gif"
cyclones = []
names = []
timestamps = []
lats = []
longs = []
basins = []
winds = []
pressures = []
tc_classes = []
lats_real = []
longs_real = []
movement_speeds = []
movement_dirs = []
log = logging.getLogger(__name__)

# increase compatibility with python<3.11
if not hasattr(datetime, "UTC"):
    datetime.UTC = datetime.timezone.utc


class ATCFError(Exception):
    """An Exception for general ATCF errors."""

    pass


class WrongData(Exception):
    """An Exception for incorrectly formatted data."""

    pass


class NoActiveStorms(Exception):
    """Used by get_forecast() to signal that no storms are active."""

    pass


async def main():
    print("CycloMonitor ATCF Module")
    print("CLI coming soon")


def reset():
    """Reset ATCF data."""
    global cyclones, names, timestamps, lats, longs, basins, winds, pressures
    global tc_classes, lats_real, longs_real, movement_speeds, movement_dirs
    cyclones.clear()
    names.clear()
    timestamps.clear()
    lats.clear()
    longs.clear()
    basins.clear()
    winds.clear()
    pressures.clear()
    tc_classes.clear()
    lats_real.clear()
    longs_real.clear()
    movement_speeds.clear()
    movement_dirs.clear()


def parse_storm(line: str, *, mode="std"):
    """Parse ATCF data.

    Arguments:
    line -- the data to be parsed
    Keyword arguments:
    mode -- parsing mode (default "std")
    If mode is "interp", parse extra data provided in ATCF's
    interp_sector_file.
    """
    log.debug(ATCF_PARSE_STORM.format(line, mode))
    storm = line.split()
    try:
        if mode == "interp":
            assert len(storm) == 12, ATCF_ERROR_COL.format(12, mode, len(storm))
        else:
            assert len(storm) == 9, ATCF_ERROR_COL.format(9, mode, len(storm))
        if not mode == "interp":
            cyclones.append(storm[0])
            names.append(storm[1])
            time = storm[2] + storm[3]
            # convert the timestamp from the given data to Unix time
            timestamp = datetime.datetime.strptime(time, "%y%m%d%H%M")
            timestamp = timestamp.replace(tzinfo=datetime.UTC)
            utc_time = int(timestamp.timestamp())
            timestamps.append(utc_time)
            lats.append(storm[4])
            longs.append(storm[5])
            basins.append(storm[6])
            winds.append(int(storm[7]))
            pressures.append(int(storm[8]))
        else:
            lats_real.append(float(storm[4]))
            longs_real.append(float(storm[5]))
            tc_classes.append(storm[7])
            movement_speeds.append(float(storm[10]))
            movement_dirs.append(float(storm[11]))
    except (AssertionError, LookupError, ValueError) as e:
        # remove the faulty entry
        faulty_entry = len(cyclones) - 1
        for index, (cy, n, ts, lat, lon, b, w, p) in enumerate(
            zip_longest(
                cyclones, names, timestamps, lats, longs, basins, winds, pressures
            )
        ):
            if index == faulty_entry:
                if cy is not None:
                    cyclones.remove(cy)
                if n is not None:
                    names.remove(n)
                if ts is not None:
                    timestamps.remove(ts)
                if lat is not None:
                    lats.remove(lat)
                if lon is not None:
                    longs.remove(lon)
                if b is not None:
                    basins.remove(b)
                if w is not None:
                    winds.remove(w)
                if p is not None:
                    pressures.remove(p)

        if mode == "interp":
            for index, (la, lo, c, ms, md) in enumerate(
                zip_longest(
                    lats_real, longs_real, tc_classes, movement_speeds, movement_dirs
                )
            ):
                if index == faulty_entry:
                    if la is not None:
                        lats_real.remove(la)
                    if lo is not None:
                        longs_real.remove(lo)
                    if c is not None:
                        tc_classes.remove(c)
                    if ms is not None:
                        movement_speeds.remove(ms)
                    if md is not None:
                        movement_dirs.remove(md)

        log.exception(ATCF_WRONG_DATA.format(line))
        raise WrongData(ATCF_WRONG_DATA.format(line)) from e


def load():
    """Load ATCF data saved on disk."""
    try:
        with open("atcf_sector_file", "r") as file:
            for line in file:
                try:
                    parse_storm(line)
                except WrongData:
                    continue
    except FileNotFoundError:
        log.info(ATCF_NO_DATA)
        return

    try:
        with open("interp_sector_file", "r") as file:
            storms = []
            for line in file:
                storms.append(line.split())
            for tc in cyclones:
                for storm in storms:
                    # sort interp data
                    cid = storm[0][2] + storm[0][3] + storm[6]
                    if cid == tc:
                        lats_real.append(float(storm[4]))
                        longs_real.append(float(storm[5]))
                        tc_classes.append(storm[7])
                        movement_speeds.append(float(storm[10]))
                        movement_dirs.append(float(storm[11]))
                        break
                else:  # no break
                    raise ATCFError(ERROR_HDYGH)
    except Exception as e:
        raise ATCFError(ATCF_GET_INTERP_FAILED) from e


# load cached data upon bringing in the module
load()


async def get_data():
    """Download ATCF data."""
    global cyclones, names, timestamps, lats, longs, basins, winds, pressures
    global tc_classes, lats_real, longs_real, movement_speeds, movement_dirs
    reset()
    log.info(ATCF_USING_MAIN)
    async with aiohttp.ClientSession(
        timeout=aiohttp.ClientTimeout(connect=10), raise_for_status=True
    ) as session:
        try:
            async with session.get(URL, ssl=False) as r:
                with open("atcf_sector_file", "w") as f:
                    f.write(await r.text())
            async with session.get(URL_INTERP, ssl=False) as r:
                with open("interp_sector_file", "w") as f:
                    f.write(await r.text())
        except Exception as e:
            log.warning(ATCF_USING_MAIN_FAILED.format(e))
            await get_data_alt()
            return
    load()
    # safeguard for some situations where the main ATCF website is down
    if not cyclones:
        await get_data_alt()


async def get_data_alt():
    """Download ATCF data (alt source)."""
    global cyclones, names, timestamps, lats, longs, basins, winds, pressures
    global tc_classes, lats_real, longs_real, movement_speeds, movement_dirs
    reset()
    log.info(ATCF_USING_ALT)
    async with aiohttp.ClientSession(
        timeout=aiohttp.ClientTimeout(connect=10), raise_for_status=True
    ) as session:
        try:
            async with session.get(URL_ALT) as r:
                tc_list = await r.json()
        except asyncio.TimeoutError as e:
            raise ATCFError(ERROR_TIMED_OUT) from e
        except aiohttp.ClientError as exc:
            raise ATCFError(ERROR_ATCF_GET_DATA_FAILED) from exc

    with open("atcf_sector_file", "w") as f:
        for d in tc_list:
            f.write(d["atcf_sector_file"] + "\n")

    with open("interp_sector_file", "w") as f:
        for d in tc_list:
            f.write(d["interp_sector_file"] + "\n")

    for d in tc_list:
        try:
            parse_storm(d["atcf_sector_file"])
            parse_storm(d["interp_sector_file"], mode="interp")
        except WrongData:
            continue


async def get_forecast(*, name="", cid=""):
    """Download the official forecast image for an active TC.
    Returns the extension of the downloaded image file.

    Keyword arguments:
    name -- Search by name
    cid -- Search by identifier
    You must specify either name or cid. name has priority over cid.
    """
    if not (cyclones and names):
        raise NoActiveStorms()
    if not (name or cid):
        raise ValueError(ERROR_GET_FORECAST_NO_PARAMS)
    if name:
        name = name.upper()
        try:
            index = names.index(name)
        except ValueError:
            index = -1
    else:
        index = -1
    if index == -1 or not name:
        try:
            index = cyclones.index(cid)
        except ValueError:
            return None

    with open("interp_sector_file", "r") as f:
        lines = f.readlines()
        atcf_id = lines[index].split()[0].upper()

    basin = atcf_id[:2]  # 2-char basin identifier
    num = atcf_id[2:4]  # 2-digit storm number
    jtwc_year = atcf_id[6:]  # Last 2 digits of the year
    if basin == "AL":
        nhc_basin = "AT"
    elif basin == "EP" or basin == "CP":
        nhc_basin = basin
    else:
        nhc_basin = None

    async with aiohttp.ClientSession(
        timeout=aiohttp.ClientTimeout(connect=10), raise_for_status=True
    ) as session:
        if nhc_basin is not None:
            nhc_id = f"{nhc_basin}{num}"
            coro = session.get(BASE_URL_NHC.format(nhc_id, atcf_id))
        else:
            basin = basin.lower()
            jtwc_id = f"{basin}{num}{jtwc_year}"
            coro = session.get(BASE_URL_JTWC.format(jtwc_id))

        async with coro as r:
            # assuming everything is good, this will either be png or gif
            ext = r.content_type.split("/")[1]
            with open(f"forecast.{ext}", "wb") as img:
                img.write(await r.read())
    return ext


if __name__ == "__main__":
    asyncio.run(main())



# see server_vars.py for documentation on related functions
# CycloMonitor Copyright (C) 2023 Nathaniel Greenwell
# This program comes with ABSOLUTELY NO WARRANTY; for details see main.py
import json
import logging

log = logging.getLogger(__name__)
json_file = "globalVars.json"


def write(var_name: str, value):
    try:
        with open(json_file, "r") as f:
            data = json.load(f)
    except FileNotFoundError:
        data = {}
    data.update({var_name: value})
    with open(json_file, "w") as f:
        f.write(json.dumps(data, indent=4))


def get(var_name: str):
    try:
        with open(json_file, "r") as f:
            data = json.load(f)
    except Exception:
        log.warning("Cannot open JSON file")
        return None
    value = data.get(var_name)
    return value



"""
CycloMonitor - ATCF and IBTrACS wrapper for Discord
Copyright (c) 2023 Virtual Nate

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
details.

For those hosting a copy of this bot, you should have received a copy of the
GNU Affero General Public License along with this program. If not, see
<https://www.gnu.org/licenses/>.
"""

from . import cli
import datetime
import logging
import asyncio
from sys import exit
from .locales import *

logname = "bot.log"
# PLEASE CHANGE THESE LINKS IF YOU ARE FORKING THIS PROJECT.
INVITE = "https://discord.com/api/oauth2/authorize?client_id={0}&permissions=67496000&scope=bot"
GITHUB = "GITHUB_LINK"
SERVER = "SERVER_INVITE"
emojis = {}

# increase compatibility with python<3.11
if not hasattr(datetime, "UTC"):
    datetime.UTC = datetime.timezone.utc

try:
    from tendo import singleton
except ImportError:
    singleton = None

try:
    from . import cyclomonitor
    from .cyclomonitor import bot
except ImportError:
    bot = None


def main():
    import argparse
    import json

    global GITHUB, INVITE, SERVER
    locale_init()
    parser = argparse.ArgumentParser(
        prog="cyclomonitor",
        description=DESCRIPTION,
        epilog=HELP_EPILOG.format(GITHUB),
    )
    parser.add_argument("-b", "--bot", help=HELP_BOT, action="store_true")
    parser.add_argument("-t", "--token", help=HELP_TOKEN, default="")
    parser.add_argument(
        "-i", "--interactive", help=HELP_INTERACTIVE, action="store_true"
    )
    parser.add_argument("-l", "--log-file", help=HELP_LOG_FILE, default="")
    parser.add_argument("-v", "--verbose", help=HELP_VERBOSE, action="store_true")
    parser.add_argument("-c", "--config", help=HELP_CONFIG, default="")
    args = parser.parse_args()
    log_params = {}

    if args.bot or args.token:
        run_bot = True
        # support the legacy TOKEN file
        try:
            with open("TOKEN", "r") as f:
                _token = f.read().split()[0]  # split in case of any newlines or spaces
        except FileNotFoundError:
            pass

        if args.token:
            _token = args.token
    else:
        run_bot = False
    if run_bot and args.interactive:
        raise ValueError(ERROR_BOT_AND_INTERACTIVE)

    if args.log_file:
        log_params["filename"] = args.log_file
        log_params["filemode"] = "a"

    if args.config:
        with open(args.config) as f:
            config = json.load(f)
        _token = config["token"]
        if config.get("log_file") is not None:
            log_params["filename"] = config["log_file"]
            log_params["filemode"] = "a"
            logging.captureWarnings(True)
        if config.get("github") is not None:
            GITHUB = config["github"]
        if config.get("client_id") is not None:
            INVITE = INVITE.format(config["client_id"])
        if config.get("server") is not None:
            SERVER = config["server"]
        if isinstance(config.get("emojis"), dict):
            emojis.update(config["emojis"])
    if args.verbose:
        log_params["level"] = logging.DEBUG
    else:
        log_params["level"] = logging.INFO
    logging.basicConfig(
        format="%(asctime)s.%(msecs)d %(name)s %(levelname)s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        **log_params,
    )
    if args.bot:
        if bot is None:
            raise ModuleNotFoundError(ERROR_PYCORD_MISSING)
        if singleton is None:
            raise ModuleNotFoundError(ERROR_TENDO_MISSING)

        cyclomonitor.GITHUB = GITHUB
        cyclomonitor.INVITE = INVITE
        cyclomonitor.SERVER = SERVER
        cyclomonitor.emojis = emojis

        try:
            # Prevent more than one instance from running at once
            me = singleton.SingleInstance()
        except singleton.SingleInstanceException:
            exit(ERROR_ALREADY_RUNNING)

        try:
            bot.run(_token)
        except NameError:
            raise ValueError(ERROR_NO_TOKEN)
        return
    if args.interactive:
        if args.config:
            logging.warning(LOG_BOT_CONFIG_AND_INTERACTIVE)
        if log_params.get("filename") is not None:
            raise ValueError(ERROR_LOG_FILE_AND_INTERACTIVE)
        cli.cli()
        return
    parser.print_help()


if __name__ == "__main__":
    main()



"""
CycloMonitor IBTrACS module

Classes:
Storm -- dataclass representing a TC
Query -- like Storm but less detailed
Generators:
query_group -- self-explanatory
Functions:
update_db -- update database
init_db -- initialize database
get_storm -- find TCs
:copyright: (c) 2024 by Nathaniel Greenwell.
"""

import sqlite3
import aiohttp
import logging
import os
import subprocess
import io
from dataclasses import dataclass
from typing import Iterable, Literal, Tuple
from .locales import *

log = logging.getLogger(__name__)
PATH = os.path.dirname(os.path.realpath(__file__))
DB = f"{PATH}/BestTrack.db"
BASE_URI = "https://www.ncei.noaa.gov/data/international-best-track-archive-for-climate-stewardship-ibtracs/v04r00/access/csv"
if not os.path.exists(DB):
    log.info(IBTRACS_DB_NOT_FOUND)


@dataclass(frozen=True)
class Storm:
    """A dataclass representing a TC.

    Attributes:
    atcf_id -- the storm's ATCF ID
    basin -- the basin the storm formed in
    peak_winds -- the storm's maximum wind speed
    peak_pres -- the storm's minimum pressure
    time_of_peak -- time the storm peaked in ISO format
    name -- the storm's name
    best_track_id -- the storm's IBTrACS ID
    season -- the year the storm formed in
    Methods:
    nature() -- cyclonic nature at peak
    is_subtropical()
    """

    atcf_id: str
    basin: str
    peak_winds: int
    peak_pres: int
    time_of_peak: str
    name: str = "NOT_NAMED"
    best_track_id: str = ""
    season: int = 0

    def nature(self) -> str:
        """Based on peak winds, return a string."""
        subtropical = self.is_subtropical()
        if self.peak_winds:
            if self.peak_winds < 34:
                if subtropical:
                    return CLASS_SD
                else:
                    return CLASS_TD
            if self.peak_winds < 64:
                if subtropical:
                    return CLASS_SS
                else:
                    return CLASS_TS
            if self.basin == "WP":
                if self.peak_winds < 130:
                    return CLASS_TY
                else:
                    return CLASS_STY
            if self.basin == "NA" or self.basin == "SA" or self.basin == "EP":
                return CLASS_HU
        if subtropical:
            return CLASS_STC
        # this value does double-duty as a generic term and the term for hurricane-strength storms in basins not listed above
        return CLASS_TC

    def is_subtropical(self, *, table="LastThreeYears"):
        """Determinte whether or not this TC was subtropical at peak."""
        if not self.peak_winds:
            return False
        for v in self.__dict__.values():
            if isinstance(v, str) and ("'" in v or ";" in v):
                return False
        con = sqlite3.connect(DB)
        cur = con.cursor()
        if self.best_track_id:
            params = [self.best_track_id, self.peak_winds, self.peak_winds]
            conds = f"SID = ? AND NATURE != 'ET' AND (WMO_WIND = ? OR USA_WIND = ?)"
        else:
            params = [
                self.name,
                self.season,
                self.basin,
                self.peak_winds,
                self.peak_winds,
            ]
            conds = f"NAME = ? AND SEASON = ? AND BASIN = ? AND NATURE != 'ET' AND (WMO_WIND = ? OR USA_WIND = ?)"
        res = cur.execute(f"SELECT NATURE FROM {table} WHERE {conds}", params)
        natures = res.fetchall()
        con.close()
        if natures:
            if (("SS",) in natures or ("DS",) in natures) and ("TS",) not in natures:
                return True
        else:
            if not natures and table == "LastThreeYears":
                return self.is_subtropical(table="AllBestTrack")
        return False


@dataclass(frozen=True, repr=False)
class Query:
    """General TC info; no peak.

    Attributes:
    sid -- IBTrACS ID
    season -- year of formation
    basin -- basin the storm formed in
    name -- the storm's name
    """

    sid: str
    season: int
    basin: str
    name: str

    def __repr__(self):
        return QUERY_REPR.format(self=self)


def query_group(queries: Iterable[Tuple[str, int, str, str]]):
    """Yield Query objects.

    Arguments:
    queries -- an iterable consisting of tuples that have length 4
    """
    for sid, season, basin, name in queries:
        yield Query(sid, season, basin, name)


def _remove_headers(csv: os.PathLike | str):
    if isinstance(csv, os.PathLike):
        csv = os.fspath(csv)
    csv_name = os.path.splitext(csv)[0]
    with open(csv) as f:
        lines = f.readlines()
        data = lines[2:]
    new_name = f"{csv_name}_NO_HEADING.csv"
    try:
        with open(new_name, "w") as f:
            f.writelines(data)
    # There has to be an except or finally clause before an else clause.
    except OSError:
        pass
    else:
        os.unlink(csv)
    finally:
        # free up RAM
        del data, lines


def _csv_import(table: Literal["LastThreeYears", "AllBestTrack"]):
    """(Internal) Import CSV into table."""
    with sqlite3.connect(DB) as con:
        cur = con.cursor()
        if table == "LastThreeYears":
            cur.executescript(open(f"{PATH}/ibtracs_LAST3.sql").read())
            filename = "ibtracs_last3_NO_HEADING.csv"
        else:
            cur.executescript(open(f"{PATH}/ibtracs_ALL.sql").read())
            filename = "ibtracs_all_NO_HEADING.csv"

        with open(f"{PATH}/{filename}") as data:
            for line in data:
                values = line.split(",")
                for i, v in enumerate(values):
                    if "." in v:
                        try:
                            values[i] = float(v)
                        except ValueError:
                            pass
                    else:
                        try:
                            values[i] = int(v)
                        except ValueError:
                            pass
                cur.execute(
                    f"INSERT INTO {table} VALUES({'?, ' * (len(values) - 1)}?)", values
                )
        con.commit()


async def update_db(mode="last3"):
    """Update the best track database.

    Arguments:
    mode -- Table(s) to update (default "last3")
    mode can be one of "last3", "all", or "full".
    If mode == "last3", update the table LastThreeYears.
    If mode == "all", update the table AllBestTrack.
    If mode == "full", update both tables.
    """
    locale_init()
    get_last3 = mode == "last3" or mode == "full"
    get_all = mode == "all" or mode == "full"
    if not (get_last3 or get_all):
        raise ValueError(ERROR_ILLEGAL_UPDATE_MODE.format(mode))
    if mode == "last3":
        log.info(IBTRACS_UPDATE_LAST3)
    elif mode == "all":
        log.info(IBTRACS_UPDATE_ALL)
    else:
        log.info(IBTRACS_UPDATE_FULL)
    log.info(IBTRACS_GETTING_DATA)
    if get_last3:
        csv = "ibtracs_last3.csv"
        async with aiohttp.ClientSession(
            timeout=aiohttp.ClientTimeout(connect=10), raise_for_status=True
        ) as session:
            try:
                r = await session.get(f"{BASE_URI}/ibtracs.last3years.list.v04r00.csv")
                with open(f"{PATH}/{csv}", "w") as f:
                    f.write(await r.text())
            except Exception:
                log.exception(ERROR_IBTRACS_UPDATE_FAILURE)
                raise
            finally:
                # Calling close does not delete the object.
                # We want to delete the resource afterwards to save RAM
                # because we may be working with a large amount of data.
                del r
        _remove_headers(f"{PATH}/{csv}")
        _csv_import("LastThreeYears")
        os.unlink(f"{PATH}/ibtracs_last3_NO_HEADING.csv")
    if get_all:
        csv = "ibtracs_all.csv"
        async with aiohttp.ClientSession(
            timeout=aiohttp.ClientTimeout(connect=10), raise_for_status=True
        ) as session:
            try:
                r = await session.get(f"{BASE_URI}/ibtracs.ALL.list.v04r00.csv")
                with open(f"{PATH}/{csv}", "w") as f:
                    f.write(await r.text())
            except Exception:
                log.exception(ERROR_IBTRACS_UPDATE_FAILURE)
                raise
            finally:
                # Calling close does not delete the object.
                # We want to delete the resource afterwards to save RAM
                # because we may be working with a large amount of data.
                del r
        _remove_headers(f"{PATH}/{csv}")
        _csv_import("AllBestTrack")
        os.unlink(f"{PATH}/ibtracs_all_NO_HEADING.csv")
    log.info(IBTRACS_UPDATE_SUCCESS)


async def init_db():
    """Equivalent to :func:`update_db("full")`"""
    await update_db("full")


def get_storm(
    *,
    name=None,
    season: int = 0,
    basin=None,
    atcf_id=None,
    ibtracs_id=None,
    table=None,
    lang="C",
):
    """Find a TC and return either a query_group() or a Storm().

    If only one storm is found, return a Storm() object.
    If more than one storm is found, return a query_group() object.
    Keyword arguments:
    name -- Filter by name (default None)
    season -- Filter by year (default 0)
    basin -- Filter by basin (default None)
    basin can be one of "NA", "SA", "NI", "SI", "SP", "EP", or "WP".
    atcf_id -- Filter by ATCF ID (default None)
    ibtracs_id -- Filter by IBTrACS ID (default None)
    table -- Set preferred database table (default "LastThreeYears")
    table can be one of "LastThreeYears" or "AllBestTrack".
    At least one of the above keyword arguments (except for table) must be
    specified by the user.
    """
    set_locale(lang)
    if not os.path.exists(DB):
        raise FileNotFoundError(ERROR_MISSING_IBTRACS_DB)
    if table is not None and table not in ["LastThreeYears", "AllBestTrack"]:
        raise ValueError(ERROR_INVALID_TABLE.format(table))
    conds_buff = io.StringIO()
    params = []
    if isinstance(basin, str):
        if basin.upper() not in ["NA", "SA", "NI", "SI", "SP", "EP", "WP"]:
            raise ValueError(ERROR_INVALID_BASIN.format(basin))
    if not isinstance(season, int):
        raise TypeError(ERROR_INVALID_SEASON)
    if name is not None:
        params.append(name.upper())
        conds_buff.write("NAME = ?")
    if season:
        if params:
            conds_buff.write(" AND ")
        params.append(season)
        conds_buff.write("SEASON = ?")
    if basin is not None:
        if params:
            conds_buff.write(" AND ")
        params.append(basin.upper())
        conds_buff.write(f"BASIN = ?")
    if atcf_id is not None:
        if params:
            conds_buff.write(" AND ")
        params.append(atcf_id.upper())
        conds_buff.write(f"USA_ATCF_ID = ?")
    if ibtracs_id is not None:
        if params:
            conds_buff.write(" AND ")
        params.append(ibtracs_id.upper())
        conds_buff.write(f"SID = ?")
    if not params:
        raise ValueError(ERROR_NO_PARAMS)
    conds = conds_buff.getvalue()
    conds_buff.close()
    con = sqlite3.connect(DB)
    cur = con.cursor()
    if table is None:
        table = "LastThreeYears"
    log.debug(IBTRACS_CONDS.format(conds))
    res = cur.execute(
        f"SELECT SID, SEASON, BASIN, NAME FROM {table} WHERE {conds}", params
    )
    data = res.fetchall()
    if not data:
        table = "AllBestTrack"
        res = cur.execute(
            f"SELECT SID, SEASON, BASIN, NAME FROM AllBestTrack WHERE {conds}", params
        )
        data = res.fetchall()
        if not data:
            return None
    storms = sorted(set(data))  # Deduplicate and sort data
    sid = storms[0][0]
    if len(storms) > 1:
        for storm in storms:
            if sid not in storm:
                con.close()
                return query_group(storms)
    params = [sid]
    res = cur.execute(
        f"SELECT USA_ATCF_ID, BASIN, MAX(USA_WIND), USA_PRES, ISO_TIME, NAME, SEASON FROM {table} WHERE SID = ? AND USA_WIND != ' ' AND NATURE != 'ET'",
        params,
    )
    atcf_id, basin, wind, pres, time, name, season = res.fetchone()
    if pres == " ":
        params = [sid, time]
        res = cur.execute(
            f"SELECT WMO_PRES FROM {table} WHERE SID = ? and ISO_TIME = ?", params
        )
        pres = res.fetchone()[0]
    if name is None:
        params = [sid]
        res = cur.execute(
            f"SELECT USA_ATCF_ID, BASIN, MAX(WMO_WIND), WMO_PRES, ISO_TIME, NAME, SEASON FROM {table} WHERE SID = ? AND WMO_WIND != ' '",
            params,
        )
        atcf_id, basin, wind, pres, time, name, season = res.fetchone()
        if name is None:
            res = cur.execute(
                f"SELECT USA_ATCF_ID, BASIN, MAX(USA_SSHS), WMO_PRES, ISO_TIME, NAME, SEASON FROM {table} WHERE SID = params"
            )
            atcf_id, basin, sshs, pres, time, name, season = res.fetchone()
            wind = 0
    if pres == " ":
        pres = 0
    if atcf_id == " ":
        atcf_id = None
    con.close()
    return Storm(atcf_id, basin, wind, pres, time, name, sid, season)


# SQLite type conversions
def varchar(val: bytes):
    return val.decode("UTF-8")


def numeric(val):
    if not val:
        return None
    else:
        return float(val)


def bit(val):
    if not val:
        return None
    else:
        return int(val)


sqlite3.register_converter("BIT", bit)
sqlite3.register_converter("VARCHAR", varchar)
sqlite3.register_converter("NUMERIC", numeric)



"""CycloMonitor Discord bot"""

import math
from . import server_vars
from . import global_vars
from . import atcf
from . import errors
import datetime
import logging
import time
import asyncio
from . import ibtracs
from types import GeneratorType
from .uptime import *
from .dir_calc import get_dir
from io import StringIO
from sys import exit
from .locales import *

copyright_notice = """
CycloMonitor - ATCF and IBTrACS wrapper for Discord
Copyright (c) 2023 Virtual Nate

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
details.

For those hosting a copy of this bot, you should have received a copy of the
GNU Affero General Public License along with this program. If not, see
<https://www.gnu.org/licenses/>.
"""

try:
    import discord
    from discord import (
        option,
        default_permissions,
        SlashCommandOptionType,
        guild_only,
        Option,
    )
    from discord.ext import tasks, commands
except ImportError as e:
    raise ModuleNotFoundError(ERROR_PYCORD_MISSING) from e

KT_TO_MPH = 1.15077945
KT_TO_KMH = 1.852
bot = discord.Bot(intents=discord.Intents.default())
# it is ideal to put out the information as soon as possible, but there may be overrides
times = [
    datetime.time(2, 0, tzinfo=datetime.UTC),
    datetime.time(8, 0, tzinfo=datetime.UTC),
    datetime.time(14, 0, tzinfo=datetime.UTC),
    datetime.time(20, 0, tzinfo=datetime.UTC),
]
log = logging.getLogger(__name__)
languages = ["C", "en_US"]
emojis = {}


class monitor(commands.Cog):
    """This class governs automated routines."""

    def __init__(self, bot):
        self.bot = bot
        self.last_update = global_vars.get("last_update")
        self.last_ibtracs_update = global_vars.get("last_ibtracs_update")
        self.is_best_track_updating = False

    def cog_unload(self):
        logging.info(LOG_MONITOR_STOP)
        self.auto_update.cancel()

    def should_suppress(self, prev_timestamps: list):
        """Compare two lists of timestamps and return a boolean."""
        suppressed = []
        for index, (cyclone, timestamp, p_timestamp) in enumerate(
            zip(atcf.cyclones, atcf.timestamps, prev_timestamps)
        ):
            try:
                # will this system request for a suppression?
                suppressed.append(p_timestamp >= timestamp)
            except IndexError:
                suppressed.append(False)
            logging.debug(LOG_TIMESTAMP_COMPARISON.format(cyclone, suppressed[index]))
        if not atcf.cyclones:
            suppressed.append(False)
        # only suppress an automatic update if all active systems requested a suppression
        for do_suppress in suppressed:
            if not do_suppress:
                return False
        return True

    @tasks.loop(time=times)
    async def auto_update(self):
        logging.info(LOG_AUTO_UPDATE_BEGIN)
        prev_timestamps = atcf.timestamps.copy()
        try:
            await atcf.get_data()
        except atcf.ATCFError as e:
            logging.exception(ERROR_AUTO_UPDATE_FAILED)
            return
        self.last_update = math.floor(time.time())
        global_vars.write("last_update", self.last_update)
        if self.should_suppress(prev_timestamps):
            # try alternate source
            logging.warn(LOG_SUPPRESSED)
            try:
                await atcf.get_data_alt()
            except atcf.ATCFError as e:
                logging.exception(ERROR_AUTO_UPDATE_FAILED)
                return
            if self.should_suppress(prev_timestamps):
                logging.warn(LOG_SUPPRESSED_TRY_2)
                for guild in bot.guilds:
                    channel_id = server_vars.get("tracking_channel", guild.id)
                    if channel_id is not None:
                        channel = bot.get_channel(channel_id)
                        await channel.send(CM_SUPPRESSED_MESSAGE)
                        await channel.send(
                            NEXT_AUTO_UPDATE.format(
                                math.floor(cog.auto_update.next_iteration.timestamp())
                            )
                        )
                return
        for guild in bot.guilds:
            channel_id = server_vars.get("tracking_channel", guild.id)
            if channel_id is not None:
                await update_guild(guild.id, channel_id)

    @auto_update.error
    async def on_update_error(self, error):
        if not isinstance(error, errors.LogRequested):
            logging.exception(CM_ERROR_WHILE_UPDATING)
        app = await self.bot.application_info()
        bot_owner = app.owner
        if bot_owner is not None:
            try:
                with open("bot.log", "rb") as log:
                    await bot_owner.send(CM_ATTACH_LOG, file=discord.File(log))
            except discord.HTTPException:
                logging.exception(ERROR_LOG_SEND_FAIL)
        else:
            logging.warning(LOG_NO_OWNER)
        if not isinstance(error, errors.LogRequested):
            for guild in bot.guilds:
                channel_id = server_vars.get("tracking_channel", guild.id)
                if channel_id is not None:
                    channel = bot.get_channel(channel_id)
                    await channel.send(CM_AUTO_UPDATE_FAILED_MESSAGE)

    @tasks.loop(time=datetime.time(0, 0, tzinfo=datetime.UTC))
    async def daily_ibtracs_update(self, *, _force_full=False):
        self.is_best_track_updating = True
        now = datetime.datetime.now(datetime.UTC)
        logging.info(LOG_IBTRACS_UPDATE_BEGIN)
        if (now.month == 1 and now.day == 2) or _force_full:
            await ibtracs.update_db("full")
        else:
            await ibtracs.update_db("last3")
        global_vars.write("last_ibtracs_update", math.floor(time.time()))
        self.is_best_track_updating = False

    @daily_ibtracs_update.error
    async def on_ibtracs_update_error(self, error):
        logging.error(LOG_IBTRACS_UPDATE_FAILED_ATTEMPT_1)
        await asyncio.sleep(10)
        for i in range(4):
            try:
                logging.info(LOG_NEXT_ATTEMPT.format(i + 2))
                await self.daily_ibtracs_update()
            except Exception as e:
                logging.error(LOG_ATTEMPT_FAILED.format(i + 2))
                if i == 3:
                    logging.exception(ERROR_IBTRACS_UPDATE_FAILED)
                    self.is_best_track_updating = False
                    raise e from error
                else:
                    logging.error(LOG_TRY_AGAIN.format(10 * (i + 2)))
                    await asyncio.sleep(10 * (i + 2))
                    continue
            else:
                break
        self.is_best_track_updating = False


async def update_guild(guild: int, to_channel: int):
    """Given a guild ID and channel ID, post ATCF data."""
    set_locale(server_vars.get("lang", guild))
    logging.info(LOG_UPDATE_GUILD.format(guild))
    channel = bot.get_channel(to_channel)
    enabled_basins = server_vars.get("basins", guild)
    current_TC_record = global_vars.get("strongest_storm")  # record-keeping
    if enabled_basins is not None:
        sent_list = []
        for (
            cyc_id,
            basin,
            wind,
            name,
            timestamp,
            lat,
            long,
            pressure,
            tc_class,
            lat_real,
            long_real,
            movement_speed,
            movement_dir,
        ) in zip(
            atcf.cyclones,
            atcf.basins,
            atcf.winds,
            atcf.names,
            atcf.timestamps,
            atcf.lats,
            atcf.longs,
            atcf.pressures,
            atcf.tc_classes,
            atcf.lats_real,
            atcf.longs_real,
            atcf.movement_speeds,
            atcf.movement_dirs,
        ):
            # per standard, we round to the nearest 5
            mph = round(wind * KT_TO_MPH / 5) * 5
            kmh = round(wind * KT_TO_KMH / 5) * 5
            c_dir = get_dir(movement_dir)
            if (not c_dir) or (movement_speed < 0):
                movement_str = NOT_AVAILABLE
            else:
                movement_mph = movement_speed * KT_TO_MPH
                movement_kph = movement_speed * KT_TO_KMH
                movement_str = STORM_MOVEMENT.format(
                    c_dir, movement_speed, movement_mph, movement_kph
                )
            # accomodate for basin crossovers
            if lat_real > 0 and long_real > 30 and long_real < 97:
                basin = "IO"
            elif lat_real > 0 and long_real > 97:
                basin = "WPAC"
            elif lat_real > 0 and long_real < -140:
                basin = "CPAC"
            elif lat_real > 0 and (
                (lat_real < 7.6 and long_real < -77)
                or (lat_real < 10 and long_real < -85)
                or (lat_real < 15 and long_real < -87)
                or (lat_real < 16 and long_real < -92.5)
                or long_real < -100
            ):
                basin = "EPAC"
            logging.debug(LOG_BASIN.format(cyc_id, basin))

            if pressure == 0:
                pressure = math.nan
            """
            Wind speeds are ignored when marking an invest.
            There is one exception, which is for subtropical cyclones, because not all agencies issue advisories/warnings on STCs (notably CPHC and JTWC).
            We can make an exception for STCs because ATCF doesn't autoflag them (more on that below).
            All wind speed values shown are in knots rounded to the nearest 5 (except for ones after > operators)
            """
            if name == "INVEST" and (not (tc_class == "SD" or tc_class == "SS")):
                tc_class = CLASS_AOI
            if tc_class == "EX":
                if not name == "INVEST":
                    tc_class = CLASS_PTC
                emoji = emojis.get("ex")
            elif tc_class == "LO" or tc_class == "INVEST":
                if not name == "INVEST":
                    tc_class = CLASS_PTC
                emoji = emojis.get("low")
            elif tc_class == "DB" or tc_class == "WV":
                if not name == "INVEST":
                    tc_class = CLASS_RL
                emoji = emojis.get("remnants")
            elif wind < 35:
                if not (tc_class == "SD" or name == "INVEST"):
                    # ignored if invest in case of autoflagging
                    # ATCF will autoflag a system to be a TD once it has attained 1-minute sustained winds of between 23 and 33 kt
                    tc_class = CLASS_TD
                    emoji = emojis.get("td")
                elif name == "INVEST" and not tc_class == "SD":
                    emoji = emojis.get("low")
                else:
                    tc_class = CLASS_SD
                    emoji = emojis.get("sd")
            elif wind > 34 and wind < 65:
                if not (tc_class == "SS" or name == "INVEST"):
                    tc_class = CLASS_TS
                    emoji = emojis.get("ts")
                elif name == "INVEST" and not tc_class == "SS":
                    emoji = emojis.get("low")
                else:
                    tc_class = CLASS_SS
                    emoji = emojis.get("ss")
            else:
                # determine the term to use based on the basin
                # we assume at this point that the system is either a TC or extratropical
                if basin == "ATL" or basin == "EPAC" or basin == "CPAC":
                    if wind < 100:
                        tc_class = CLASS_HU
                    else:
                        tc_class = CLASS_MH
                elif basin == "WPAC":
                    if wind < 130:
                        tc_class = CLASS_TY
                    else:
                        tc_class = CLASS_STY
                else:
                    tc_class = CLASS_CY

                # for custom emoji to work, the bot needs to be in the server it's from
                # you also need the emoji's ID
                if wind < 85:
                    emoji = emojis.get("cat1")
                elif wind > 84 and wind < 100:
                    emoji = emojis.get("cat2")
                elif wind > 99 and wind < 115:
                    emoji = emojis.get("cat3")
                elif wind > 114 and wind < 140:
                    emoji = emojis.get("cat4")
                elif wind > 139 and wind < 155:
                    emoji = emojis.get("cat5")
                elif wind > 154 and wind < 170:
                    emoji = emojis.get("cat5intense")
                else:
                    emoji = emojis.get("cat5veryintense")
            if name == "INVEST":
                name = display_name = cyc_id
            else:
                display_name = f"{cyc_id} ({name})"
            if emoji is None:
                emoji = ":cyclone:"
            # update TC records
            if current_TC_record is not None:
                if (wind > int(current_TC_record[5])) or (
                    wind == int(current_TC_record[5])
                    and pressure < int(current_TC_record[8])
                ):
                    logging.info(LOG_NEW_RECORD)
                    global_vars.write(
                        "strongest_storm",
                        [
                            emoji,
                            tc_class,
                            cyc_id,
                            name,
                            str(timestamp),
                            str(wind),
                            str(mph),
                            str(kmh),
                            str(pressure),
                        ],
                    )
            else:
                logging.info(LOG_NO_RECORD)
                global_vars.write(
                    "strongest_storm",
                    [
                        emoji,
                        tc_class,
                        cyc_id,
                        name,
                        str(timestamp),
                        str(wind),
                        str(mph),
                        str(kmh),
                        str(pressure),
                    ],
                )

            # this check is really long since it needs to accomodate for every possible situation
            send_message = (
                (basin == "ATL" and enabled_basins[0] == "1")
                or (basin == "EPAC" and enabled_basins[1] == "1")
                or (basin == "CPAC" and enabled_basins[2] == "1")
                or (basin == "WPAC" and enabled_basins[3] == "1")
                or (basin == "IO" and enabled_basins[4] == "1")
                or (basin == "SHEM" and enabled_basins[5] == "1")
            )
            sent_list.append(send_message)
            if math.isnan(pressure):
                pressure = "N/A"
            if send_message:
                try:
                    await channel.send(
                        CM_STORM_INFO.format(
                            emoji,
                            tc_class,
                            display_name,
                            f"<t:{timestamp}:f>",
                            name,
                            lat,
                            long,
                            wind,
                            mph,
                            kmh,
                            pressure,
                            movement_str,
                        )
                    )
                except discord.errors.HTTPException:
                    logging.warning(LOG_GUILD_UNAVAILABLE.format(guild))
                    return

        for was_sent in sent_list:
            if was_sent:
                break
        else:  # no break
            await channel.send(CM_NO_STORMS)
        try:
            next_run = int(cog.auto_update.next_iteration.timestamp())
        except AttributeError:
            next_run = NO_AUTO_UPDATE
        # it is best practice to use official sources when possible
        await channel.send(CM_NEXT_AUTO_UPDATE.format(next_run))
        await channel.send(CM_MORE_INFO)


@bot.event
async def on_ready():
    locale_init()
    await bot.change_presence(
        activity=discord.Activity(type=discord.ActivityType.watching, name=CM_WATCHING)
    )
    logging.info(LOG_READY.format(bot.user))
    global_vars.write("guild_count", len(bot.guilds))
    global cog
    # stop the monitor if it is already running
    # this is to prevent a situation where there are two instances of a task running in some edge cases
    try:
        cog.auto_update.cancel()
        cog.daily_ibtracs_update().cancel()
    except NameError:
        cog = monitor(bot)

    # keep trying to start the auto-update tasks until they have started
    while cog.auto_update.next_iteration is None:
        cog.auto_update.start()
        await asyncio.sleep(1)
    while cog.daily_ibtracs_update.next_iteration is None:
        cog.daily_ibtracs_update.start()
        await asyncio.sleep(1)
    # force an automatic update if last_update is not set or more than 6 hours have passed since the last update
    if (cog.last_update is None) or (math.floor(time.time()) - cog.last_update > 21600):
        await cog.auto_update()
    # same idea as above but for IBTrACS data, and the limit is 24 hours
    if (cog.last_ibtracs_update is None) or (
        math.floor(time.time()) - cog.last_ibtracs_update > 86400
    ):
        await cog.daily_ibtracs_update()


@bot.event
async def on_guild_join(guild: discord.Guild):
    logging.info(LOG_NEW_GUILD.format(guild.name))
    count = len(bot.guilds)
    global_vars.write("guild_count", count)
    for channel in guild.text_channels:
        if channel.permissions_for(guild.me).send_messages:
            await channel.send(CM_GUILD_ADDED)
            break


@bot.event
async def on_guild_remove(guild: discord.Guild):
    logging.info(LOG_GUILD_REMOVED.format(guild.name))
    count = len(bot.guilds)
    server_vars.remove_guild(guild.id)
    global_vars.write("guild_count", count)


@bot.event
async def on_application_command(ctx: discord.ApplicationContext):
    server_lang = server_vars.get("lang", ctx.guild_id)
    if server_lang is not None:
        set_locale(server_lang)
    else:
        set_locale("C")


@bot.event
async def on_application_command_error(ctx: discord.ApplicationContext, error):
    if isinstance(error, commands.errors.MissingPermissions) or isinstance(
        error, commands.errors.NotOwner
    ):
        try:
            await ctx.respond(CM_NO_PERMISSION, ephemeral=True)
        except discord.errors.HTTPException:
            logging.exception(ERROR_CANNOT_RESPOND)
        logging.warn(LOG_NO_PERMISSION.format(ctx.author, ctx.command.name))
    elif isinstance(error, commands.errors.NoPrivateMessage):
        try:
            await ctx.respond(CM_NO_DM, ephemeral=True)
        except discord.errors.HTTPException:
            logging.exception(ERROR_CANNOT_RESPOND)
    else:
        logging.exception(LOG_COMMAND_ERROR.format(ctx.command.name, error))
        try:
            await ctx.respond(CM_COMMAND_ERROR.format(error), ephemeral=True)
        except discord.errors.HTTPException:
            logging.exception(ERROR_CANNOT_RESPOND)


@bot.slash_command(name="ping", description=CM_PING)
async def ping(ctx):
    await ctx.defer(ephemeral=True)
    await ctx.respond(CM_PONG.format(int(bot.latency * 1000)), ephemeral=True)


@bot.slash_command(name="set_tracking_channel", description=CM_SET_TRACKING_CHANNEL)
@guild_only()
@commands.has_guild_permissions(manage_channels=True)
@option("channel", discord.TextChannel, description=CM_CHANNEL_TO_USE)
async def set_tracking_channel(ctx, channel):
    await ctx.defer(ephemeral=True)
    if not isinstance(channel, discord.TextChannel):
        await ctx.respond(ERROR_NOT_A_TEXT_CHANNEL, ephemeral=True)
    else:
        if channel.permissions_for(ctx.me).send_messages:
            server_vars.write("tracking_channel", channel.id, ctx.guild_id)
            await ctx.respond(CM_SET_CHANNEL_SUCCESS.format(channel), ephemeral=True)
        else:
            await ctx.respond(CM_CANNOT_SEND_MESSAGE, ephemeral=True)


@bot.slash_command(name="update", description=CM_UPDATE)
@guild_only()
@commands.has_guild_permissions(manage_messages=True)
async def update(ctx):
    await ctx.defer(ephemeral=True)
    channel_id = server_vars.get("tracking_channel", ctx.guild_id)
    await atcf.get_data()
    cog.last_update = math.floor(time.time())
    global_vars.write("last_update", cog.last_update)
    if channel_id is not None:
        await update_guild(ctx.guild_id, channel_id)
        await ctx.respond(CM_UPDATE_SUCCESS, ephemeral=True)
    else:
        await ctx.respond(ERROR_NO_TRACKING_CHANNEL, ephemeral=True)
    atcf.reset()


@bot.slash_command(name="update_alt", description=CM_UPDATE_ALT)
@guild_only()
@commands.has_guild_permissions(manage_messages=True)
async def update_alt(ctx):
    await ctx.defer(ephemeral=True)
    channel_id = server_vars.get("tracking_channel", ctx.guild_id)
    await atcf.get_data_alt()
    cog.last_update = math.floor(time.time())
    global_vars.write("last_update", cog.last_update)
    if channel_id is not None:
        await update_guild(ctx.guild_id, channel_id)
        await ctx.respond(CM_UPDATE_SUCCESS, ephemeral=True)
    else:
        await ctx.respond(ERROR_NO_TRACKING_CHANNEL, ephemeral=True)
    atcf.reset()


@bot.slash_command(name="set_basins", description=CM_SET_BASINS)
@guild_only()
@commands.has_guild_permissions(manage_guild=True)
async def set_basins(
    ctx: discord.ApplicationContext,
    natl: Option(bool, CM_NATL),
    epac: Option(bool, CM_EPAC),
    cpac: Option(bool, CM_CPAC),
    wpac: Option(bool, CM_WPAC),
    nio: Option(bool, CM_NIO),
    shem: Option(bool, CM_SHEM),
):
    await ctx.defer(ephemeral=True)
    # fmt: off
    # this effectively represents a 6-bit binary value
    enabled_basins = f"{int(natl)}{int(epac)}{int(cpac)}{int(wpac)}{int(nio)}{int(shem)}"
    # fmt: on
    server_vars.write("basins", enabled_basins, ctx.guild_id)
    await ctx.respond(CM_BASINS_SAVED, ephemeral=True)


@bot.slash_command(name="update_all", description=CM_UPDATE_ALL)
@commands.is_owner()
async def update_all(ctx):
    await ctx.defer(ephemeral=True)
    await atcf.get_data()
    cog.last_update = math.floor(time.time())
    global_vars.write("last_update", cog.last_update)
    for guild in bot.guilds:
        channel_id = server_vars.get("tracking_channel", guild.id)
        # attempt to update only if the tracking channel is set
        if channel_id is not None:
            await update_guild(guild.id, channel_id)
    await ctx.respond(CM_UPDATE_SUCCESS, ephemeral=True)


@bot.slash_command(name="update_all_alt", description=CM_UPDATE_ALL_ALT)
@commands.is_owner()
async def update_all_alt(ctx):
    await ctx.defer(ephemeral=True)
    await atcf.get_data_alt()
    cog.last_update = math.floor(time.time())
    global_vars.write("last_update", cog.last_update)
    for guild in bot.guilds:
        channel_id = server_vars.get("tracking_channel", guild.id)
        # attempt to update only if the tracking channel is set
        if channel_id is not None:
            await update_guild(guild.id, channel_id)
    await ctx.respond(CM_UPDATE_SUCCESS, ephemeral=True)


@bot.slash_command(name="announce_all", description=CM_ANNOUNCE_ALL)
@commands.is_owner()
async def announce_all(
    ctx: discord.ApplicationContext, announcement: Option(str, CM_TO_ANNOUNCE)
):
    await ctx.defer(ephemeral=True)
    for guild in bot.guilds:
        channel_id = server_vars.get("tracking_channel", guild.id)
        if channel_id is not None:
            channel = bot.get_channel(channel_id)
            await channel.send(announcement)
    await ctx.respond(CM_ANNOUNCE_ALL_SUCCESS.format(announcement), ephemeral=True)


@bot.slash_command(name="announce_basin", description=CM_ANNOUNCE_BASIN)
@commands.is_owner()
async def announce_basin(
    ctx: discord.ApplicationContext,
    basin: Option(
        str,
        CM_BASIN_TO_ANNOUNCE,
        choices=["natl", "epac", "cpac", "wpac", "nio", "shem"],
    ),
    announcement: Option(str, CM_TO_ANNOUNCE),
):
    await ctx.defer(ephemeral=True)
    for guild in bot.guilds:
        channel_id = server_vars.get("tracking_channel", guild.id)
        enabled_basins = server_vars.get("basins", guild.id)
        if channel_id is not None:
            channel = bot.get_channel(channel_id)
            send_message = (
                (basin == "natl" and enabled_basins[0] == "1")
                or (basin == "epac" and enabled_basins[1] == "1")
                or (basin == "cpac" and enabled_basins[2] == "1")
                or (basin == "wpac" and enabled_basins[3] == "1")
                or (basin == "nio" and enabled_basins[4] == "1")
                or (basin == "shem" and enabled_basins[5] == "1")
            )
            if send_message:
                await channel.send(CM_BASIN_ANNOUNCEMENT.format(basin, announcement))
    await ctx.respond(
        CM_ANNOUNCE_BASIN_SUCCESS.format(basin, announcement), ephemeral=True
    )


@bot.slash_command(name="announce_file", description=CM_ANNOUNCE_FILE)
@commands.is_owner()
async def announce_file(
    ctx: discord.ApplicationContext,
    file: Option(discord.SlashCommandOptionType.attachment, CM_TXT_FILE),
):
    await ctx.defer(ephemeral=True)
    print(file.content_type)
    if file.content_type.startswith("text"):
        with open("announcement.md", "wb") as f:
            await file.save(f)
        announcement = open("announcement.md", "r").read()
        for guild in bot.guilds:
            channel_id = server_vars.get("tracking_channel", guild.id)
            if channel_id is not None:
                channel = bot.get_channel(channel_id)
                await channel.send(announcement)
        await ctx.respond(CM_ANNOUNCE_ALL_SUCCESS.format(announcement), ephemeral=True)
    else:
        await ctx.respond(ERROR_NOT_A_TXT_FILE, ephemeral=True)


@bot.slash_command(name="invite", description=CM_INVITE)
async def invite(ctx):
    await ctx.respond(CM_INVITE_MESSAGE.format(INVITE), ephemeral=True)


@bot.slash_command(name="statistics", description=CM_STATISTICS_DESC)
async def statistics(ctx):
    await ctx.defer()
    strongest_storm = global_vars.get("strongest_storm")
    if strongest_storm is None:
        # If the above method returned None then it means that it cannot load the JSON file.
        await ctx.respond(ERROR_NO_GLOBAL_VARS, ephemeral=True)
        return
    yikes_count = global_vars.get("yikes_count")
    guild_count = global_vars.get("guild_count")
    await ctx.respond(
        CM_STATISTICS.format(
            guild_count, strongest_storm, yikes_count, process_uptime_human_readable()
        )
    )


@bot.slash_command(name="yikes", description=CM_YIKES)
async def yikes(ctx):
    await ctx.defer(ephemeral=True)
    count = global_vars.get("yikes_count")
    if count is not None:
        count += 1
    else:
        count = 1
    global_vars.write("yikes_count", count)
    for guild in bot.guilds:
        channel_id = server_vars.get("tracking_channel", guild.id)
        if channel_id is not None:
            channel = bot.get_channel(channel_id)
            await channel.send(CM_INC_YIKES_COUNT)
    logging.info(CM_INC_YIKES_COUNT)
    await ctx.respond(CM_YIKES_RESPONSE)


@bot.slash_command(name="get_data", description=CM_GET_DATA)
@commands.is_owner()
async def get_data(ctx):
    await ctx.defer(ephemeral=True)
    try:
        await atcf.get_data()
        cog.last_update = math.floor(time.time())
        global_vars.write("last_update", cog.last_update)
        with open("atcf_sector_file", "r") as f:
            content = f.read()
            await ctx.respond(CM_GET_DATA_SUCCESS.format(content), ephemeral=True)
    except atcf.ATCFError as e:
        await ctx.respond(CM_GET_DATA_FAILED.format(e), ephemeral=True)


@bot.slash_command(name="get_data_alt", description=CM_GET_DATA_ALT)
@commands.is_owner()
async def get_data_alt(ctx):
    await ctx.defer(ephemeral=True)
    try:
        await atcf.get_data_alt()
        cog.last_update = math.floor(time.time())
        global_vars.write("last_update", cog.last_update)
        with open("atcf_sector_file", "r") as f:
            content = f.read()
            await ctx.respond(CM_GET_DATA_SUCCESS.format(content), ephemeral=True)
    except atcf.ATCFError as e:
        await ctx.respond(CM_GET_DATA_FAILED.format(e), ephemeral=True)


@bot.slash_command(name="atcf_reset", description=CM_ATCF_RESET)
@commands.is_owner()
async def atcf_reset(ctx):
    atcf.reset()
    await ctx.respond(CM_ATCF_RESET_SUCCESS, ephemeral=True)


@bot.slash_command(name="github", description=CM_GITHUB)
async def github(ctx):
    await ctx.respond(CM_GITHUB_RESPONSE.format(GITHUB), ephemeral=True)


@bot.slash_command(name="copyright", description=CM_COPYRIGHT)
async def copyright(ctx):
    await ctx.respond(copyright_notice, ephemeral=True)


@bot.slash_command(name="rsmc_list", description=CM_RSMC_LIST)
async def rsmc_list(ctx):
    await ctx.defer(ephemeral=True)
    await ctx.respond(CM_RSMC_LIST_RESPONSE)


@bot.slash_command(name="get_log", description=CM_GET_LOG)
@commands.is_owner()
async def get_log(ctx):
    await ctx.defer(ephemeral=True)
    await cog.on_update_error(errors.LogRequested(LOG_REQUESTED))
    await ctx.respond(CM_LOG_SENT)


@bot.slash_command(name="suspend_updates", description=CM_SUSPEND_UPDATES)
@commands.is_owner()
async def suspend_updates(ctx):
    await ctx.defer(ephemeral=True)
    if cog.auto_update.next_iteration is not None:
        cog.auto_update.cancel()
        await ctx.respond(CM_SUSPEND_UPDATES_SUCCESS)
    else:
        await ctx.respond(CM_UPDATES_ALREADY_SUSPENDED)


@bot.slash_command(name="resume_updates", description=CM_RESUME_UPDATES)
@commands.is_owner()
async def resume_updates(ctx):
    await ctx.defer(ephemeral=True)
    if cog.auto_update.next_iteration is None:
        cog.auto_update.start()
        await ctx.respond(CM_RESUME_UPDATES_SUCCESS)
    else:
        await ctx.respond(CM_UPDATES_ALREADY_RUNNING)


@bot.slash_command(name="feedback", description=CM_FEEDBACK)
async def feedback(
    ctx: discord.ApplicationContext, msg: Option(str, CM_FEEDBACK_TO_SEND)
):
    await ctx.defer(ephemeral=True)
    app = await bot.application_info()
    bot_owner = app.owner
    if bot_owner is not None:
        await bot_owner.send(CM_FEEDBACK_RECEIVED.format(ctx.author, msg))
        await ctx.respond(CM_FEEDBACK_SENT.format(msg))
    else:
        await ctx.respond(CM_NO_OWNER)


@bot.slash_command(name="get_past_storm", description=CM_GET_PAST_STORM)
async def get_past_storm(
    ctx: discord.ApplicationContext,
    name: Option(str, CM_PAST_STORM_NAME, default=None),
    season: Option(int, CM_PAST_STORM_SEASON, min_value=1841, default=0),
    basin: Option(
        str,
        CM_PAST_STORM_BASIN,
        choices=["NA", "SA", "EP", "WP", "SP", "NI", "SI"],
        default=None,
    ),
    atcf_id: Option(str, CM_PAST_STORM_ATCF, default=None),
    ibtracs_id: Option(str, CM_PAST_STORM_SID, default=None),
    table: Option(
        str,
        CM_PAST_STORM_TABLE,
        choices=["LastThreeYears", "AllBestTrack"],
        default="LastThreeYears",
    ),
):
    await ctx.defer(ephemeral=True)
    if cog.is_best_track_updating:
        await ctx.respond(CM_WAIT_FOR_IBTRACS_UPDATE)
        response = await ctx.interaction.original_response()
        while cog.is_best_track_updating:
            await asyncio.sleep(0.25)
        await response.edit_message(content=CM_SEARCHING)
    else:
        await ctx.respond(CM_SEARCHING)
        response = await ctx.interaction.original_response()
    try:
        results = ibtracs.get_storm(
            name=name,
            season=season,
            basin=basin,
            atcf_id=atcf_id,
            ibtracs_id=ibtracs_id,
            table=table,
            lang=server_vars.get("lang", ctx.guild_id),
        )
    except ValueError as e:
        await response.edit(CM_ERROR.format(e))
        return
    if isinstance(results, GeneratorType):
        res_tmp = StringIO()
        res_tmp.write(CM_MULTIPLE_STORMS)
        lines = [f"{s}\n" for s in results]
        res_tmp.writelines(lines)
        await response.edit(res_tmp.getvalue())
        res_tmp.close()
    elif isinstance(results, ibtracs.Storm):
        peak_timestamp = int(
            datetime.datetime.fromisoformat(results.time_of_peak)
            .replace(tzinfo=datetime.UTC)
            .timestamp()
        )
        nature = results.nature().title()
        name = results.name.title()
        if name == "Not_Named":
            descriptor = CM_UNNAMED_STORM.format(nature)
        else:
            descriptor = f"{nature} {name}"
        if not results.peak_winds:
            peak_winds = CM_UNKNOWN
            peak_time = CM_UNKNOWN
        else:
            peak_winds = f"{results.peak_winds} kt"
            peak_time = f"<t:{peak_timestamp}:f>"
        if not results.peak_pres:
            peak_pres = CM_UNKNOWN
        else:
            peak_pres = f"{results.peak_pres} mb"
        if results.atcf_id is None:
            atcf_id = CM_UNKNOWN
        else:
            atcf_id = results.atcf_id
        await response.edit(
            content=CM_PAST_STORM_INFO.format(
                descriptor,
                results.season,
                results.basin,
                peak_winds,
                peak_pres,
                peak_time,
                atcf_id,
                results.best_track_id,
            )
        )
    elif results is None:
        await response.edit(content=CM_NO_RESULTS)
    else:
        # you should never see this error ;-)
        raise TypeError(ERROR_WTF.format(type(results)))


@bot.slash_command(name="set_language", description=CM_SET_LANGUAGE)
async def set_language(
    ctx: discord.ApplicationContext,
    language: Option(str, CM_LANG_TO_USE, choices=languages),
):
    await ctx.defer(ephemeral=True)
    server_vars.write("lang", language, ctx.guild.id)
    set_locale(language)
    await ctx.respond(CM_SET_LANGUAGE_SUCCESS.format(language))


@bot.slash_command(name="get_forecast", description=CM_GET_FORECAST)
async def get_forecast(
    ctx: discord.ApplicationContext,
    name: Option(str),
):
    await ctx.defer()
    name = name.upper()
    if name == "INVEST":
        await ctx.respond(CM_IS_AN_INVEST)
        return

    try:
        ext = await atcf.get_forecast(name=name)
    except atcf.NoActiveStorms:
        await ctx.respond(CM_NO_ACTIVE_STORMS)
    except Exception as e:
        await on_application_command_error(ctx, e)
    else:
        if ext is None:
            await ctx.respond(
                CM_CANNOT_FIND_STORM.format(
                    "\n".join([n for n in atcf.names if n != "INVEST"])
                )
            )
        else:
            with open(f"forecast.{ext}", "rb") as f:
                await ctx.respond(file=discord.File(f))


@bot.slash_command(name="server", description=CM_SERVER)
async def server(ctx: discord.ApplicationContext):
    await ctx.respond(SERVER, ephemeral=True)



# CycloMonitor Copyright (C) 2023 Nathaniel Greenwell
# This program comes with ABSOLUTELY NO WARRANTY; for details see main.py
import json
import logging

log = logging.getLogger(__name__)
json_file = "serverVars.json"


def write(var_name: str, value, guild: int):
    # if the json file exists, load it, otherwise initialize a blank list
    try:
        with open(json_file, "r") as f:
            data = json.load(f)
    except FileNotFoundError:
        data = []
    for i in data:
        if i.get(str(guild)) is not None:
            server_data = i.get(str(guild))
            server_data.update({var_name: value})
            break
    # this code segment is run if the loop was not broken
    else:
        data.append({str(guild): {var_name: value}})
    if len(data) == 0:
        data.append({str(guild): {var_name: value}})
    with open(json_file, "w") as f:
        f.write(json.dumps(data, indent=4))


# this is expected to be assigned to a variable, so we return None if no data can be loaded
def get(var_name: str, guild: int):
    try:
        with open(json_file, "r") as f:
            data = json.load(f)
    except Exception:
        log.warning("Cannot open JSON file")
        return None
    for i in data:
        if i.get(str(guild)) is not None:
            server_data = i.get(str(guild))
            value = server_data.get(var_name)
            return value
    return None


def remove_guild(guild: int):
    try:
        with open(json_file, "r") as f:
            data = json.load(f)
    except Exception:
        log.warning("Cannot open JSON file")
        return None
    for i in data:
        if i.get(str(guild)) is not None:
            data.remove(i)
            with open(json_file, "w") as f:
                f.write(json.dumps(data, indent=4))



from setuptools import setup, find_packages
import pathlib

__version__ = "2024.5.15"

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name="cyclomonitor",
    version=__version__,
    description="Discord bot that helps you keep tabs on tropical cyclones.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ntvmb/cyclomonitor",
    author="Nathaniel Greenwell",
    author_email="nategreenwell@live.com",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Internet",
        "Topic :: Scientific/Engineering",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3 :: Only",
    ],
    keywords=[
        "weather",
        "discord",
        "discord bot",
        "discord-py",
        "typhoon",
        "hurricane",
        "tropical cyclone",
        "atcf",
        "py-cord",
        "ibtracs",
    ],
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3.8",
    install_requires=["aiohttp>=3.6.0"],
    extras_require={
        "bot": ["py-cord>=2.5.0", "tendo>=0.3.0"],
    },
    project_urls={
        "Bug Reports": "https://github.com/ntvmb/cyclomonitor/issues",
        "Source": "https://github.com/ntvmb/cyclomonitor",
    },
)
